import { id, map, ap, prepend, reduce, thunk } from '../../vendor/izuna/src/index.js'
/** @import { Morphism } from './types.js' */


/**
 * @template A
 * @typedef {Pure<A> | Impure<A>} Free
 */

/** @template A */
export class Pure {
  #value

  /**
   * @param {A} value 
   */
  constructor(value) {
    this.#value = value
  }

  /**
   * @template {Free<B>} B
   * @param {Morphism<A, B>} f
   * @returns {B}
   */
  chain(f) {
    return f(this.#value)
  }

  /**
   * @template B
   * @param {Morphism<A, B>} f
   * @returns {Free<B>}
   */
  map(f) {
    // @ts-ignore
    return pure(f(this.#value))
  }

  /**
   * @template B, C
   * @param {Free<B>} b 
   * @returns {Free<C>}
   */
  ap(b) {
    return b.map(this.#value)
  }

  traverse(f, of) {
    return of(this.map(f))
  }

  /**
   * @template B
   * @param {(acc: B, value: A) => B} f 
   * @param {B} init
   * @returns {B}
   */
  reduce(f, init) {
    return f(init, this.#value)
  }

}

/**
 * @template A
 * @typedef {<B>(value: A) => Free<B>} Computation
 */

/** @template A */
export class Impure {
  #next

  /**
   * @param {() => Free<any>} next 
   */
  constructor(next) {
    this.#next = next
  }

  /**
   * @template {Free<B>} B
   * @param {Morphism<A, B>} f
   * @returns {B}
   */
  chain(f) {
    // @ts-ignore
    return new Impure(
      // @ts-ignore
      () => this.#next().chain(f)
    )
  }

  /**
   * @template B
   * @param {Morphism<A, B>} f
   * @returns {Free<B>}
   */
  map(f) {
    return new Impure(
      // @ts-ignore
      () => this.#next().map(f)
    )
  }

  /**
   * @template B
   * @param {Free<any>} b 
   * @returns {Free<B>}
   */
  ap(b) {
    return new Impure(
      // @ts-ignore
      () => b.map(this.#next())
    )
  }

  traverse(f, of) {
    return of(liftF(
      this.#next().traverse(f, of)
    ))
  }

  /**
   * @template B
   * @param {(acc: B, value: A) => B} f 
   * @param {B} acc
   * @returns {B}
   */
  reduce(f, acc) {

  }
}

const sequence = (of, iter) => {
  return reduce((acc, x) => {
    ap(acc, map(prepend, x))
  }, of([]), iter)
}

export const pure = x => new Pure(x)
export const impure = f => new Impure(f)

export const liftF = effect => impure(thunk(effect))

